/****************************************************************************/
/**                                                                        **/
/**                          Connect-4 Algorithm                           **/
/**                                                                        **/
/**                              Version 3.4                               **/
/**                                                                        **/
/**                            By Keith Pomakis                            **/
/**                          (pomakis@pobox.com)                           **/
/**                                                                        **/
/**                              April, 1997                               **/
/**                                                                        **/
/**                  ported by thomas.nussbaumer@gmx.net                   **/
/**                                                                        **/
/**                               June, 2000                               **/
/**                                                                        **/
/**                                                                        **/
/****************************************************************************/
/**                                                                        **/
/**  This file provides the functions necessary to implement a front-end-  **/
/**  independent, device-independent Connect-4 game.  Multiple board sizes **/
/**  are supported.  It is also possible to specify the number of pieces   **/
/**  necessary to connect in a row in order to win.  Therefore one can     **/
/**  play Connect-3, Connect-5, etc.  An efficient tree-searching          **/
/**  algorithm (making use of alpha-beta cutoff decisions) has been        **/
/**  implemented to insure that the computer plays as quickly as possible. **/
/**                                                                        **/
/**  The declaration of the public functions necessary to use this file    **/
/**  are contained in "c4algo.h".                                          **/
/**                                                                        **/
/**  In all of the public functions (all of which have the "c4_" prefix),  **/
/**  the value of player can be any integer, where an even integer refers  **/
/**  to player 0 and an odd integer refers to player 1.                    **/
/**                                                                        **/
/****************************************************************************/

#include "c4algo.h"

/* Some macros for convenience. */

#define other(x)        ((x) ^ 1)
#define real_player(x)  ((x) & 1)

#define pop_state() \
        (current_state = &state_stack[--depth])

/* The "goodness" of the current state with respect to a player is the */
/* score of that player minus the score of the player's opponent.  A   */
/* positive value will result if the specified player is in a better   */
/* situation than his/her opponent.                                    */

#define goodness_of(player) \
        (current_state->score[player] - current_state->score[other(player)])

/* A local struct which defines the state of a game. */

typedef struct {

    char **board;           /* The board configuration of the game state.  */
                            /* board[x][y] specifies the position of the   */
                            /* xth column and the yth row of the board,    */
                            /* where column and row numbering starts at 0. */
                            /* (The 0th row is the bottom row.)            */
                            /* A value of 0 specifies that the position is */
                            /* occupied by a piece owned by player 0, a    */
                            /* value of 1 specifies that the position is   */
                            /* occupied by a piece owned by player 1, and  */
                            /* a value of C4_NONE specifies that the       */
                            /* position is unoccupied.                     */

    short *(score_array[2]);  /* An array specifying statistics on both      */
                            /* players.  score_array[0] specifies the      */
                            /* statistics for player 0, while              */
                            /* score_array[1] specifies the statistics for */
                            /* player 1.                                   */

    short score[2];           /* The actual scores of each player, deducible */
                            /* from score_array, but kept separately for   */
                            /* efficiency.  The score of player x is the   */
                            /* sum of score_array[x].  A score is          */
                            /* basically a function of how many winning    */
                            /* positions are still available to the        */
                            /* and how close he/she is to achieving each   */
                            /* of these positions.                         */

    short winner;       /* The winner of the game - either 0, 1 or     */
                            /* C4_NONE.  Deducible from score_array, but   */
                            /* kept separately for efficiency.             */

    short num_of_pieces;      /* The number of pieces currently occupying    */
                            /* board spaces.  Deducible from board, but    */
                            /* kept separately for efficiency.             */

} Game_state;

/* Static global variables. */

short       size_x           = 0;
short       size_y           = 0;
short       num_to_connect   = 0;
short       win_places       = 0;
short       magic_win_number = 0;
short       depth            = 0;
short      *drop_order       = NULL;
Boolean  ***map              = NULL;
Boolean     game_in_progress = FALSE;
Boolean     move_in_progress = FALSE;
Boolean     break_game       = FALSE;
Game_state  state_stack[C4_MAX_LEVEL+1];
Game_state *current_state    = NULL;


/* A declaration of the local functions. */

static inline short num_of_win_places(short x, short y, short n);
static inline void  update_score(short player, short x, short y);
static short drop_piece(short player, short column);
static void  push_state(void);
static inline short evaluate(short player, short level, short alpha, short beta);


/*************************************************************************/
/* (NOT PART OF ORIGINAL ENGINE)                                         */
/* frees allocated game memory                                           */
/*************************************************************************/
static inline void FreeGameMemory() {
    short       i;
    Game_state *tmp;

    for (i=0;i<C4_MAX_LEVEL+1;i++) {
        tmp = &(state_stack[i]);
        if (tmp->board) {
            if (tmp->board[0]) free(tmp->board[0]);
            free(tmp->board);
            tmp->board = NULL;
        }
        else {
            break;
        }
    }

    if (drop_order) {
        free(drop_order);
        drop_order = NULL;
    }

    if (map) {
        for (i=0; i<MAX_BOARD_WIDTH; i++) {
            if (map[i]) {
                if (map[i][0]) free(map[i][0]);
                free(map[i]);
                map[i] = NULL;
            }
            else {
                break;
            }
        }
        free(map);
        map = NULL;
    }
}


/*************************************************************************/
/* (NOT PART OF ORIGINAL ENGINE)                                         */
/* to make the game memory usage more predictable we allocate everything */
/* at once                                                               */
/* returns 1 if everything went okay                                     */
/*************************************************************************/
static inline short AllocGameMemory() {
    Game_state *tmp;
    short       i,j;
    short       wplaces = num_of_win_places(MAX_BOARD_WIDTH,MAX_BOARD_HEIGHT,MAX_BOARD_CONNECT);

    // initialize all pointers with NULL
    for (i=0;i<C4_MAX_LEVEL+1;i++) {
        state_stack[i].board = NULL;
    }
    drop_order = NULL;
    map        = NULL;

    // allocate memory for complete stack
    for (i=0;i<C4_MAX_LEVEL+1;i++) {
        tmp = &(state_stack[i]);

        if (!(tmp->board    = (char **)malloc(MAX_BOARD_WIDTH * sizeof(char *)))) return 0;
        if (!(tmp->board[0] = (char *)malloc(MAX_BOARD_WIDTH*MAX_BOARD_HEIGHT+wplaces*sizeof(short)*2+8))) return 0;

        for (j=1; j<MAX_BOARD_WIDTH;j++) {
             tmp->board[j] = tmp->board[0]+MAX_BOARD_HEIGHT*j;
        }

        tmp->score_array[0] = (short *) (tmp->board[0]+MAX_BOARD_HEIGHT*MAX_BOARD_WIDTH);
        tmp->score_array[1] = (short *) tmp->score_array[0]+wplaces;
    }

    // allocate space for drop_order
    if (!(drop_order = (short *)malloc(MAX_BOARD_WIDTH * sizeof(short)))) return 0;

    // allocate space for map
    if (!(map = (Boolean ***)malloc(MAX_BOARD_WIDTH * sizeof(Boolean **)))) return 0;

    for (i=0; i<MAX_BOARD_WIDTH; i++) {
        if (!(map[i] = (Boolean **)malloc(MAX_BOARD_HEIGHT * sizeof(Boolean *)))) return 0;
        if (!(map[i][0] = (Boolean *)malloc(wplaces * sizeof(Boolean) * MAX_BOARD_HEIGHT))) return 0;

        for (j=1; j<MAX_BOARD_HEIGHT; j++) {
            map[i][j] = (Boolean*)((unsigned char*)(map[i][0]) + wplaces*sizeof(Boolean)*j);
        }
    }

    return 1;
}


/****************************************************************************/
/**                                                                        **/
/**  This function sets up a new game.  This must be called exactly once   **/
/**  before each game is started.  Before it can be called a second time,  **/
/**  end_game() must be called to destroy the previous game.               **/
/**                                                                        **/
/**  width and height are the desired dimensions of the game board, while  **/
/**  num is the number of pieces required to connect in a row in order to  **/
/**  win the game.                                                         **/
/**                                                                        **/
/****************************************************************************/
static inline void c4_new_game(short width, short height, short num) {
    register short i, j, k;
    short count, column;

    size_x           = width;
    size_y           = height;
    num_to_connect   = num;
    magic_win_number = 1 << num_to_connect;
    win_places       = num_of_win_places(size_x, size_y, num_to_connect);

    /* Set up the board */
    depth         = 0;
    current_state = &state_stack[0];

    //current_state->board = (char **) emalloc(size_x * sizeof(char *));
    for (i=0; i<size_x; i++) {
        for (j=0; j<size_y; j++) current_state->board[i][j] = C4_NONE;
    }

    /* Set up the score array */
    for (i=0; i<win_places; i++) {
        current_state->score_array[0][i] = 1;
        current_state->score_array[1][i] = 1;
    }

    current_state->score[0]      = current_state->score[1] = win_places;
    current_state->winner        = C4_NONE;
    current_state->num_of_pieces = 0;


    /* Set up the map */
    for (i=0; i<size_x; i++) {
        for (j=0; j<size_y; j++) {
            for (k=0; k<win_places; k++)
                map[i][j][k] = FALSE;
        }
    }

    count = 0;

    /* Fill in the horizontal win positions */
    for (i=0; i<size_y; i++) {
        for (j=0; j<size_x-num_to_connect+1; j++) {
            for (k=0; k<num_to_connect; k++) {
                map[j+k][i][count] = TRUE;
            }
            count++;
        }
    }

    /* Fill in the vertical win positions */
    for (i=0; i<size_x; i++) {
        for (j=0; j<size_y-num_to_connect+1; j++) {
            for (k=0; k<num_to_connect; k++) {
                map[i][j+k][count] = TRUE;
            }
            count++;
        }
    }

    /* Fill in the forward diagonal win positions */
    for (i=0; i<size_y-num_to_connect+1; i++) {
        for (j=0; j<size_x-num_to_connect+1; j++) {
            for (k=0; k<num_to_connect; k++) {
                map[j+k][i+k][count] = TRUE;
            }
            count++;
        }
    }

    /* Fill in the backward diagonal win positions */
    for (i=0; i<size_y-num_to_connect+1; i++) {
        for (j=size_x-1; j>=num_to_connect-1; j--) {
            for (k=0; k<num_to_connect; k++) {
                map[j-k][i+k][count] = TRUE;
            }
            count++;
        }
    }

    /* Set up the order in which automatic moves should be tried. */
    /* The columns nearer to the center of the board are usually  */
    /* better tactically and are more likely to lead to a win.    */
    /* By ordering the search such that the central columns are   */
    /* tried first, alpha-beta cutoff is much more effective.     */

    //drop_order = (short *) emalloc(size_x * sizeof(short ));
    column = (size_x-1) / 2;
    for (i=1; i<=size_x; i++) {
        drop_order[i-1] = column;
        column += ((i%2)? i : -i);
    }

    game_in_progress = TRUE;
    break_game       = FALSE;
}


/****************************************************************************/
/**                                                                        **/
/**  This function drops a piece of the specified player into the          **/
/**  specified column.  A value of TRUE is returned if the drop is         **/
/**  successful, or FALSE otherwise.  A drop is unsuccessful if the        **/
/**  specified column number is invalid or full.  If the drop is           **/
/**  successful and row is a non-NULL pointer, the row where the piece     **/
/**  ended up is returned through the row pointer.  Note that column and   **/
/**  row numbering start at 0.                                             **/
/**                                                                        **/
/****************************************************************************/

static inline Boolean c4_make_move(short player, short column, short *row) {
    short result;

    if (column >= size_x || column < 0) return FALSE;
    result = drop_piece(real_player(player), column);
    if (row && result >= 0) *row = result;
    return (result >= 0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the computer to make a move for the specified **/
/**  player.  level specifies the number of levels deep the computer       **/
/**  should search the game tree in order to make its decision.  This      **/
/**  corresponds to the number of "moves" in the game, where each player's **/
/**  turn is considered a move.  A value of TRUE is returned if a move was **/
/**  made, or FALSE otherwise (i.e. if the board is full).  If a move was  **/
/**  made, the column and row where the piece ended up is returned through **/
/**  the column and row pointers (unless a pointer is NULL, in which case  **/
/**  it won't be used to return any information).  Note that column and    **/
/**  row numbering start at 0.  Also note that for a standard 7x6 game of  **/
/**  Connect-4, the computer is brain-dead at levels of three or less,     **/
/**  while at levels of four or more the computer provides a challenge.    **/
/**                                                                        **/
/****************************************************************************/

static inline Boolean c4_auto_move(short player, short level, short *column, short *row) {
    short i, best_column = -1, goodness = 0, best_worst = -(INT_MAX);
    short num_of_equal = 0, real_player, current_column, result;

    move_in_progress = TRUE;
    real_player = real_player(player);

    /* Simulate a drop in each of the columns and see what the results are. */
    for (i=0; i<size_x; i++) {
        push_state();
        current_column = drop_order[i];

        /* If this column is full, ignore it as a possibility. */
        if (drop_piece(real_player, current_column) < 0) {
            pop_state();
            continue;
        }

        /* If this drop wins the game, take it! */
        if (current_state->winner == real_player) {
            best_column = current_column;
            pop_state();
            break;
        }

        /* Otherwise, look ahead to see how good this move may turn out */
        /* to be (assuming the opponent makes the best moves possible). */
        else {
            goodness = evaluate(real_player, level, -(INT_MAX), -best_worst);
        }

        /* If this move looks better than the ones previously considered, */
        /* remember it.                                                   */
        if (goodness > best_worst) {
            best_worst = goodness;
            best_column = current_column;
            num_of_equal = 1;
        }

        /* If two moves are equally as good, make a random decision. */
        else if (goodness == best_worst) {
            num_of_equal++;
            if (random(num_of_equal)==0) best_column = current_column;
        }

        pop_state();
        if (break_game) break;
    }

    move_in_progress = FALSE;

    /* Drop the piece in the column decided upon. */
    if (best_column >= 0) {
        result = drop_piece(real_player, best_column);
        if (column) *column = best_column;
        if (row)    *row = result;
        return TRUE;
    }

    return FALSE;
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns a two-dimensional array containing the state of **/
/**  the game board.  Do not modify this array.  It is assumed that a game **/
/**  is in progress.  The value of this array is dynamic and will change   **/
/**  to reflect the state of the game as the game progresses.  It becomes  **/
/**  and stays undefined when the game ends.                               **/
/**                                                                        **/
/**  The first dimension specifies the column number and the second        **/
/**  dimension specifies the row number, where column and row numbering    **/
/**  start at 0 and the bottow row is considered the 0th row.  A value of  **/
/**  0 specifies that the position is occupied by a piece owned by player  **/
/**  0, a value of 1 specifies that the position is occupied by a piece    **/
/**  owned by player 1, and a value of C4_NONE specifies that the position **/
/**  is unoccupied.                                                        **/
/**                                                                        **/
/****************************************************************************/

static inline char ** c4_board(void) {
    return current_state->board;
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns TRUE if the specified player has won the game,  **/
/**  and FALSE otherwise.                                                  **/
/**                                                                        **/
/****************************************************************************/

static inline Boolean c4_is_winner(short player) {
    return (current_state->winner == real_player(player));
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns TRUE if the board is completely full, and FALSE **/
/**  otherwise.                                                            **/
/**                                                                        **/
/****************************************************************************/

static inline Boolean c4_is_tie() {
    return (current_state->num_of_pieces == size_x * size_y);
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns the coordinates of the winning connections of   **/
/**  the winning player.  It is assumed that a player has indeed won the   **/
/**  game.  The coordinates are returned in x1, y1, x2, y2, where (x1, y1) **/
/**  specifies the lower-left piece of the winning connection, and         **/
/**  (x2, y2) specifies the upper-right piece of the winning connection.   **/
/**  If more than one winning connection exists, only one will be          **/
/**  returned.                                                             **/
/**                                                                        **/
/****************************************************************************/

static inline void c4_win_coords(short *x1, short *y1, short *x2, short *y2) {
    register short i, j;
    short winner, win_pos = 0;
    Boolean found;

    winner = current_state->winner;

    while (current_state->score_array[winner][win_pos] != magic_win_number) win_pos++;

    /* Find the lower-left piece of the winning connection. */
    found = FALSE;
    for (j=0; j<size_y && !found; j++) {
        for (i=0; i<size_x; i++) {
            if (map[i][j][win_pos]) {
                *x1 = i;
                *y1 = j;
                found = TRUE;
                break;
            }
        }
    }

    /* Find the upper-right piece of the winning connection. */
    found = FALSE;
    for (j=size_y-1; j>=0 && !found; j--) {
        for (i=size_x-1; i>=0; i--) {
            if (map[i][j][win_pos]) {
                *x2 = i;
                *y2 = j;
                found = TRUE;
                break;
            }
        }
    }
}


/****************************************************************************/
/****************************************************************************/
/**                                                                        **/
/**  The following functions are local to this file and should not be      **/
/**  called externally.                                                    **/
/**                                                                        **/
/****************************************************************************/
/****************************************************************************/


/****************************************************************************/
/**                                                                        **/
/**  This function returns the number of possible win positions on a board **/
/**  of dimensions x by y with n being the number of pieces required in a  **/
/**  row in order to win.                                                  **/
/**                                                                        **/
/****************************************************************************/

static inline short num_of_win_places(short x, short y, short n) {
    if (x < n && y < n) return 0;
    else if (x < n)     return x * ((y-n)+1);
    else if (y < n)     return y * ((x-n)+1);
    else                return 4*x*y - 3*x*n - 3*y*n + 3*x + 3*y - 4*n + 2*n*n + 2;
}


/****************************************************************************/
/**                                                                        **/
/**  This function updates the score of the specified player in the        **/
/**  context of the current state,  given that the player has just placed  **/
/**  a game piece in column x, row y.                                      **/
/**                                                                        **/
/****************************************************************************/
static inline void update_score(short player, short x, short y) {
    short i;
    short this_difference = 0, other_difference = 0;
    short **current_score_array = current_state->score_array;
    short other_player = other(player);

    for (i=0; i<win_places; i++) {
        if (map[x][y][i]) {
            this_difference  += current_score_array[player][i];
            other_difference += current_score_array[other_player][i];

            current_score_array[player][i] <<= 1;
            current_score_array[other_player][i] = 0;

            this_difference -= current_score_array[player][i];

            if (current_score_array[player][i] == magic_win_number) {
                if (current_state->winner == C4_NONE) {
                    current_state->winner = player;
                }
            }
        }
    }
    current_state->score[player]       -= this_difference;
    current_state->score[other_player] -= other_difference;
}


/****************************************************************************/
/**                                                                        **/
/**  This function drops a piece of the specified player into the          **/
/**  specified column.  The row where the piece ended up is returned, or   **/
/**  -1 if the drop was unsuccessful (i.e., the specified column is full). **/
/**                                                                        **/
/****************************************************************************/
static short drop_piece(short player, short column) {
    short y = 0;

    while (current_state->board[column][y] != C4_NONE && ++y < size_y);

    if (y == size_y) return -1;

    current_state->board[column][y] = player;
    current_state->num_of_pieces++;
    update_score(player, column, y);

    return y;
}


/****************************************************************************/
/**                                                                        **/
/**  This function pushes the current state onto a stack.  popdir() is     **/
/**  used to pop from this stack.                                          **/
/**                                                                        **/
/**  Technically what it does, since the current state is considered to    **/
/**  be the top of the stack, is push a copy of the current state onto     **/
/**  the stack right above it.  The stack pointer (depth) is then          **/
/**  incremented so that the new copy is considered to be the current      **/
/**  state.  That way, all pop_state() has to do is decrement the stack    **/
/**  pointer.                                                              **/
/**                                                                        **/
/**  For efficiency, memory for each stack state used is only allocated    **/
/**  once per game, and reused for the remainder of the game.              **/
/**                                                                        **/
/****************************************************************************/

static void push_state(void) {
    register short i, win_places_array_size;
    Game_state *old_state, *new_state;

    win_places_array_size = win_places * sizeof(short );
    old_state = &state_stack[depth++];
    new_state = &state_stack[depth];

    /* Copy the board */
    for (i=0; i<size_x; i++) memcpy(new_state->board[i], old_state->board[i], size_y);

    /* Copy the score array */
    memcpy(new_state->score_array[0], old_state->score_array[0],win_places_array_size);
    memcpy(new_state->score_array[1], old_state->score_array[1],win_places_array_size);

    new_state->score[0] = old_state->score[0];
    new_state->score[1] = old_state->score[1];
    new_state->winner   = old_state->winner;

    current_state = new_state;
}


/****************************************************************************/
/**                                                                        **/
/**  This recursive function determines how good the current state may     **/
/**  turn out to be for the specified player.  It does this by looking     **/
/**  ahead level moves.  It is assumed that both the specified player and  **/
/**  the opponent may make the best move possible.  alpha and beta are     **/
/**  used for alpha-beta cutoff so that the game tree can be pruned to     **/
/**  avoid searching unneccessary paths.                                   **/
/**                                                                        **/
/**  The worst goodness that the current state can produce in the number   **/
/**  of moves (levels) searched is returned.  This is the best the         **/
/**  specified player can hope to achieve with this state (since it is     **/
/**  assumed that the opponent will make the best moves possible).         **/
/**                                                                        **/
/****************************************************************************/

short skipcounter = 0;

static inline short evaluate(short player, short level, short alpha, short beta) {
    short i, goodness, best, maxab;

    //--------------------------------------------------
    // special treatment for TI's to make break possible
    // skipping is done for speedup
    //--------------------------------------------------
    if (skipcounter>=200) {
        skipcounter=0;
        if (kbhit()) {
            if (GetUserInput() == KEY_ESC) {
                if (RequestDialog("Sure to leave Game ?")) {
                    break_game = TRUE;
                    return 0;
                }
            }
        }
    }
    skipcounter++;

    if (level == depth) return goodness_of(player);
    else {
        /* Assume it is the other player's turn. */
        best = -(INT_MAX);
        maxab = alpha;
        for(i=0; i<size_x; i++) {
            push_state();
            if (drop_piece(other(player), drop_order[i]) < 0) {
                pop_state();
                continue;
            }
            if (current_state->winner == other(player)) goodness = INT_MAX - depth;
            else                                        goodness = evaluate(other(player), level, -beta, -maxab);

            if (goodness > best) {
                best = goodness;
                if (best > maxab) maxab = best;
            }
            pop_state();
            if (best > beta || break_game) break;
        }

        /* What's good for the other player is bad for this one. */
        return -best;
    }
}


